"""
If you use this code, please cite one of the SynthSeg papers:
https://github.com/BBillot/SynthSeg/blob/master/bibtex.bib

Copyright 2020 Benjamin Billot

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software distributed under the License is
distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
implied. See the License for the specific language governing permissions and limitations under the
License.
"""

# python imports
import os
import numpy as np
import tensorflow as tf
from keras import models
from keras import layers as KL

# project imports
from lamareg.SynthSeg import metrics_model as metrics
from lamareg.SynthSeg.training import train_model
from lamareg.SynthSeg.labels_to_image_model import get_shapes
from lamareg.SynthSeg.training_supervised import build_model_inputs

# third-party imports
from lamareg.ext.lab2im import utils, layers
from lamareg.ext.neuron import models as nrn_models


def training(
    list_paths_input_labels,
    list_paths_target_labels,
    model_dir,
    input_segmentation_labels,
    target_segmentation_labels=None,
    subjects_prob=None,
    batchsize=1,
    output_shape=None,
    scaling_bounds=0.2,
    rotation_bounds=15,
    shearing_bounds=0.012,
    nonlin_std=3.0,
    nonlin_scale=0.04,
    prob_erosion_dilation=0.3,
    min_erosion_dilation=4,
    max_erosion_dilation=5,
    n_levels=5,
    nb_conv_per_level=2,
    conv_size=5,
    unet_feat_count=16,
    feat_multiplier=2,
    activation="elu",
    skip_n_concatenations=2,
    lr=1e-4,
    wl2_epochs=1,
    dice_epochs=50,
    steps_per_epoch=10000,
    checkpoint=None,
):
    """

    This function trains a UNet to segment MRI images with synthetic scans generated by sampling a GMM conditioned on
    label maps. We regroup the parameters in four categories: General, Augmentation, Architecture, Training.

    # IMPORTANT !!!
    # Each time we provide a parameter with separate values for each axis (e.g. with a numpy array or a sequence),
    # these values refer to the RAS axes.

    :param list_paths_input_labels: list of all the paths of the input label maps. These correspond to "noisy"
    segmentations that the denoiser will be trained to correct.
    :param list_paths_target_labels: list of all the paths of the output label maps. Must have the same order as
    list_paths_input_labels. These are the target label maps that the network will learn to produce given the "noisy"
    input label maps.
    :param model_dir: path of a directory where the models will be saved during training.
    :param input_segmentation_labels: list of all the label values present in the input label maps.
    :param target_segmentation_labels: list of all the label values present in the output label maps. By default (None)
    this will be taken to be the same as input_segmentation_labels.

    # ----------------------------------------------- General parameters -----------------------------------------------
    # label maps parameters
    :param subjects_prob: (optional) relative order of importance (doesn't have to be probabilistic), with which to pick
    the provided label maps at each minibatch. Can be a sequence, a 1D numpy array, or the path to such an array, and it
    must be as long as path_label_maps. By default, all label maps are chosen with the same importance.

    # output-related parameters
    :param batchsize: (optional) number of images to generate per mini-batch. Default is 1.
    :param output_shape: (optional) desired shape of the output image, obtained by randomly cropping the generated image
    Can be an integer (same size in all dimensions), a sequence, a 1d numpy array, or the path to a 1d numpy array.
    Default is None, where no cropping is performed.

    # --------------------------------------------- Augmentation parameters --------------------------------------------
    # spatial deformation parameters
    :param scaling_bounds: (optional) if apply_linear_trans is True, the scaling factor for each dimension is
    sampled from a uniform distribution of predefined bounds. Can either be:
    1) a number, in which case the scaling factor is independently sampled from the uniform distribution of bounds
    (1-scaling_bounds, 1+scaling_bounds) for each dimension.
    2) the path to a numpy array of shape (2, n_dims), in which case the scaling factor in dimension i is sampled from
    the uniform distribution of bounds (scaling_bounds[0, i], scaling_bounds[1, i]) for the i-th dimension.
    3) False, in which case scaling is completely turned off.
    Default is scaling_bounds = 0.2 (case 1)
    :param rotation_bounds: (optional) same as scaling bounds but for the rotation angle, except that for case 1 the
    bounds are centred on 0 rather than 1, i.e. (0+rotation_bounds[i], 0-rotation_bounds[i]).
    Default is rotation_bounds = 15.
    :param shearing_bounds: (optional) same as scaling bounds. Default is shearing_bounds = 0.012.
    :param nonlin_std: (optional) Standard deviation of the normal distribution from which we sample the first
    tensor for synthesising the deformation field. Set to 0 to completely deactivate elastic deformation.
    :param nonlin_scale: (optional) Ratio between the size of the input label maps and the size of the sampled
    tensor for synthesising the elastic deformation field.

    # degradation of the input labels
    :param prob_erosion_dilation: (optional) probability with which to degrade the input label maps with erosion or
    dilation. If 0, then no erosion/dilation is applied to the label maps given as inputs to the network.
    :param min_erosion_dilation: (optional) when prob_erosion_dilation is not zero, erosion and dilation of random
    coefficients are applied. Set the minimum erosion/dilation coefficient here.
    :param max_erosion_dilation: (optional) Set the maximum erosion/dilation coefficient here.

    # ------------------------------------------ UNet architecture parameters ------------------------------------------
    :param n_levels: (optional) number of level for the Unet. Default is 5.
    :param nb_conv_per_level: (optional) number of convolutional layers per level. Default is 2.
    :param conv_size: (optional) size of the convolution kernels. Default is 2.
    :param unet_feat_count: (optional) number of feature for the first layer of the UNet. Default is 24.
    :param feat_multiplier: (optional) multiply the number of feature by this number at each new level. Default is 2.
    :param activation: (optional) activation function. Can be 'elu', 'relu'.
    :param skip_n_concatenations: (optional) number of levels for which to remove the traditional skip connections of
    the UNet architecture. default is zero, which corresponds to the classic UNet architecture. Example:
    If skip_n_concatenations = 2, then we will remove the concatenation link between the two top levels of the UNet.

    # ----------------------------------------------- Training parameters ----------------------------------------------
    :param lr: (optional) learning rate for the training. Default is 1e-4
    :param wl2_epochs: (optional) number of epochs for which the network (except the soft-max layer) is trained with L2
    norm loss function. Default is 1.
    :param dice_epochs: (optional) number of epochs with the soft Dice loss function. Default is 50.
    :param steps_per_epoch: (optional) number of steps per epoch. Default is 10000. Since no online validation is
    possible, this is equivalent to the frequency at which the models are saved.
    :param checkpoint: (optional) path of an already saved model to load before starting the training.
    """

    # check epochs
    assert (wl2_epochs > 0) | (
        dice_epochs > 0
    ), "either wl2_epochs or dice_epochs must be positive, had {0} and {1}".format(
        wl2_epochs, dice_epochs
    )

    # prepare data files
    input_label_list, _ = utils.get_list_labels(label_list=input_segmentation_labels)
    if target_segmentation_labels is None:
        target_label_list = input_label_list
    else:
        target_label_list, _ = utils.get_list_labels(
            label_list=target_segmentation_labels
        )
    n_labels = np.size(target_label_list)

    # create augmentation model
    labels_shape, _, _, _, _, _ = utils.get_volume_info(
        list_paths_input_labels[0], aff_ref=np.eye(4)
    )
    augmentation_model = build_augmentation_model(
        labels_shape,
        input_label_list,
        crop_shape=output_shape,
        output_div_by_n=2**n_levels,
        scaling_bounds=scaling_bounds,
        rotation_bounds=rotation_bounds,
        shearing_bounds=shearing_bounds,
        nonlin_std=nonlin_std,
        nonlin_scale=nonlin_scale,
        prob_erosion_dilation=prob_erosion_dilation,
        min_erosion_dilation=min_erosion_dilation,
        max_erosion_dilation=max_erosion_dilation,
    )
    unet_input_shape = augmentation_model.output[0].get_shape().as_list()[1:]

    # prepare the segmentation model
    l2l_model = nrn_models.unet(
        input_model=augmentation_model,
        input_shape=unet_input_shape,
        nb_labels=n_labels,
        nb_levels=n_levels,
        nb_conv_per_level=nb_conv_per_level,
        conv_size=conv_size,
        nb_features=unet_feat_count,
        feat_mult=feat_multiplier,
        activation=activation,
        batch_norm=-1,
        skip_n_concatenations=skip_n_concatenations,
        name="l2l",
    )

    # input generator
    model_inputs = build_model_inputs(
        path_inputs=list_paths_input_labels,
        path_outputs=list_paths_target_labels,
        batchsize=batchsize,
        subjects_prob=subjects_prob,
        dtype_input="int32",
    )
    input_generator = utils.build_training_generator(model_inputs, batchsize)

    # pre-training with weighted L2, input is fit to the softmax rather than the probabilities
    if wl2_epochs > 0:
        wl2_model = models.Model(
            l2l_model.inputs, [l2l_model.get_layer("l2l_likelihood").output]
        )
        wl2_model = metrics.metrics_model(wl2_model, target_label_list, "wl2")
        train_model(
            wl2_model,
            input_generator,
            lr,
            wl2_epochs,
            steps_per_epoch,
            model_dir,
            "wl2",
            checkpoint,
        )
        checkpoint = os.path.join(model_dir, "wl2_%03d.h5" % wl2_epochs)

    # fine-tuning with dice metric
    dice_model = metrics.metrics_model(l2l_model, target_label_list, "dice")
    train_model(
        dice_model,
        input_generator,
        lr,
        dice_epochs,
        steps_per_epoch,
        model_dir,
        "dice",
        checkpoint,
    )


def build_augmentation_model(
    labels_shape,
    segmentation_labels,
    crop_shape=None,
    output_div_by_n=None,
    scaling_bounds=0.15,
    rotation_bounds=15,
    shearing_bounds=0.012,
    translation_bounds=False,
    nonlin_std=3.0,
    nonlin_scale=0.0625,
    prob_erosion_dilation=0.3,
    min_erosion_dilation=4,
    max_erosion_dilation=7,
):

    # reformat resolutions and get shapes
    labels_shape = utils.reformat_to_list(labels_shape)
    n_dims, _ = utils.get_dims(labels_shape)
    n_labels = len(segmentation_labels)

    # get shapes
    crop_shape, _ = get_shapes(
        labels_shape,
        crop_shape,
        np.array([1] * n_dims),
        np.array([1] * n_dims),
        output_div_by_n,
    )

    # define model inputs
    net_input = KL.Input(
        shape=labels_shape + [1], name="l2l_noisy_labels_input", dtype="int32"
    )
    target_input = KL.Input(
        shape=labels_shape + [1], name="l2l_target_input", dtype="int32"
    )

    # deform labels
    noisy_labels, target = layers.RandomSpatialDeformation(
        scaling_bounds=scaling_bounds,
        rotation_bounds=rotation_bounds,
        shearing_bounds=shearing_bounds,
        translation_bounds=translation_bounds,
        nonlin_std=nonlin_std,
        nonlin_scale=nonlin_scale,
        inter_method="nearest",
    )([net_input, target_input])

    # cropping
    if crop_shape != labels_shape:
        noisy_labels, target = layers.RandomCrop(crop_shape)([noisy_labels, target])

    # random erosion
    if prob_erosion_dilation > 0:
        noisy_labels = layers.RandomDilationErosion(
            min_erosion_dilation, max_erosion_dilation, prob=prob_erosion_dilation
        )(noisy_labels)

    # convert input labels (i.e. noisy_labels) to [0, ... N-1] and make them one-hot
    noisy_labels = layers.ConvertLabels(np.unique(segmentation_labels))(noisy_labels)
    target = KL.Lambda(lambda x: tf.cast(x[..., 0], "int32"), name="labels_out")(target)
    noisy_labels = KL.Lambda(
        lambda x: tf.one_hot(x[0][..., 0], depth=n_labels), name="noisy_labels_out"
    )([noisy_labels, target])

    # build model and return
    brain_model = models.Model(
        inputs=[net_input, target_input], outputs=[noisy_labels, target]
    )
    return brain_model



"""
If you use this code, please cite one of the SynthSeg papers:
https://github.com/BBillot/SynthSeg/blob/master/bibtex.bib

Copyright 2020 Benjamin Billot

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software distributed under the License is
distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
implied. See the License for the specific language governing permissions and limitations under the
License.
"""

# python imports
import numpy as np
import numpy.random as npr

# third-party imports
from lamareg.ext.lab2im import utils


def build_model_inputs(
    path_label_maps,
    n_labels,
    batchsize=1,
    n_channels=1,
    subjects_prob=None,
    generation_classes=None,
    prior_distributions="uniform",
    prior_means=None,
    prior_stds=None,
    use_specific_stats_for_channel=False,
    mix_prior_and_random=False,
):
    """
    This function builds a generator that will be used to give the necessary inputs to the label_to_image model: the
    input label maps, as well as the means and stds defining the parameters of the GMM (which change at each minibatch).
    :param path_label_maps: list of the paths of the input label maps.
    :param n_labels: number of labels in the input label maps.
    :param batchsize: (optional) numbers of images to generate per mini-batch. Default is 1.
    :param n_channels: (optional) number of channels to be synthesised. Default is 1.
    :param subjects_prob: (optional) relative order of importance (doesn't have to be probabilistic), with which to pick
    the provided label maps at each minibatch. Must be a 1D numpy array, as long as path_label_maps.
    :param generation_classes: (optional) Indices regrouping generation labels into classes of same intensity
    distribution. Regrouped labels will thus share the same Gaussian when sampling a new image. Can be a sequence or a
    1d numpy array. It should have the same length as generation_labels, and contain values between 0 and K-1, where K
    is the total number of classes. Default is all labels have different classes.
    :param prior_distributions: (optional) type of distribution from which we sample the GMM parameters.
    Can either be 'uniform', or 'normal'. Default is 'uniform'.
    :param prior_means: (optional) hyperparameters controlling the prior distributions of the GMM means. Because
    these prior distributions are uniform or normal, they require by 2 hyperparameters. Thus prior_means can be:
    1) a sequence of length 2, directly defining the two hyperparameters: [min, max] if prior_distributions is
    uniform, [mean, std] if the distribution is normal. The GMM means of are independently sampled at each
    mini_batch from the same distribution.
    2) an array of shape (2, K), where K is the number of classes (K=len(generation_labels) if generation_classes is
    not given). The mean of the Gaussian distribution associated to class k in [0, ...K-1] is sampled at each mini-batch
    from U(prior_means[0,k], prior_means[1,k]) if prior_distributions is uniform, or from
    N(prior_means[0,k], prior_means[1,k]) if prior_distributions is normal.
    3) an array of shape (2*n_mod, K), where each block of two rows is associated to hyperparameters derived
    from different modalities. In this case, if use_specific_stats_for_channel is False, we first randomly select a
    modality from the n_mod possibilities, and we sample the GMM means like in 2).
    If use_specific_stats_for_channel is True, each block of two rows correspond to a different channel
    (n_mod=n_channels), thus we select the corresponding block to each channel rather than randomly drawing it.
    4) the path to such a numpy array.
    Default is None, which corresponds to prior_means = [25, 225].
    :param prior_stds: (optional) same as prior_means but for the standard deviations of the GMM.
    Default is None, which corresponds to prior_stds = [5, 25].
    :param use_specific_stats_for_channel: (optional) whether the i-th block of two rows in the prior arrays must be
    only used to generate the i-th channel. If True, n_mod should be equal to n_channels. Default is False.
    :param mix_prior_and_random: (optional) if prior_means is not None, enables to reset the priors to their default
    values for half of these cases, and thus generate images of random contrast.
    """

    # allocate unique class to each label if generation classes is not given
    if generation_classes is None:
        generation_classes = np.arange(n_labels)
    n_classes = len(np.unique(generation_classes))

    # make sure subjects_prob sums to 1
    subjects_prob = utils.load_array_if_path(subjects_prob)
    if subjects_prob is not None:
        subjects_prob /= np.sum(subjects_prob)

    # Generate!
    while True:

        # randomly pick as many images as batchsize
        indices = npr.choice(
            np.arange(len(path_label_maps)), size=batchsize, p=subjects_prob
        )

        # initialise input lists
        list_label_maps = []
        list_means = []
        list_stds = []

        for idx in indices:

            # load input label map
            lab = utils.load_volume(
                path_label_maps[idx], dtype="int", aff_ref=np.eye(4)
            )
            if (npr.uniform() > 0.7) & ("seg_cerebral" in path_label_maps[idx]):
                lab[lab == 24] = 0

            # add label map to inputs
            list_label_maps.append(utils.add_axis(lab, axis=[0, -1]))

            # add means and standard deviations to inputs
            means = np.empty((1, n_labels, 0))
            stds = np.empty((1, n_labels, 0))
            for channel in range(n_channels):

                # retrieve channel specific stats if necessary
                if isinstance(prior_means, np.ndarray):
                    if (prior_means.shape[0] > 2) & use_specific_stats_for_channel:
                        if prior_means.shape[0] / 2 != n_channels:
                            raise ValueError(
                                "the number of blocks in prior_means does not match n_channels. This "
                                "message is printed because use_specific_stats_for_channel is True."
                            )
                        tmp_prior_means = prior_means[2 * channel : 2 * channel + 2, :]
                    else:
                        tmp_prior_means = prior_means
                else:
                    tmp_prior_means = prior_means
                if (
                    (prior_means is not None)
                    & mix_prior_and_random
                    & (npr.uniform() > 0.5)
                ):
                    tmp_prior_means = None
                if isinstance(prior_stds, np.ndarray):
                    if (prior_stds.shape[0] > 2) & use_specific_stats_for_channel:
                        if prior_stds.shape[0] / 2 != n_channels:
                            raise ValueError(
                                "the number of blocks in prior_stds does not match n_channels. This "
                                "message is printed because use_specific_stats_for_channel is True."
                            )
                        tmp_prior_stds = prior_stds[2 * channel : 2 * channel + 2, :]
                    else:
                        tmp_prior_stds = prior_stds
                else:
                    tmp_prior_stds = prior_stds
                if (
                    (prior_stds is not None)
                    & mix_prior_and_random
                    & (npr.uniform() > 0.5)
                ):
                    tmp_prior_stds = None

                # draw means and std devs from priors
                tmp_classes_means = utils.draw_value_from_distribution(
                    tmp_prior_means,
                    n_classes,
                    prior_distributions,
                    125.0,
                    125.0,
                    positive_only=True,
                )
                tmp_classes_stds = utils.draw_value_from_distribution(
                    tmp_prior_stds,
                    n_classes,
                    prior_distributions,
                    15.0,
                    15.0,
                    positive_only=True,
                )
                random_coef = npr.uniform()
                if random_coef > 0.95:  # reset the background to 0 in 5% of cases
                    tmp_classes_means[0] = 0
                    tmp_classes_stds[0] = 0
                elif (
                    random_coef > 0.7
                ):  # reset the background to low Gaussian in 25% of cases
                    tmp_classes_means[0] = npr.uniform(0, 15)
                    tmp_classes_stds[0] = npr.uniform(0, 5)
                tmp_means = utils.add_axis(
                    tmp_classes_means[generation_classes], axis=[0, -1]
                )
                tmp_stds = utils.add_axis(
                    tmp_classes_stds[generation_classes], axis=[0, -1]
                )
                means = np.concatenate([means, tmp_means], axis=-1)
                stds = np.concatenate([stds, tmp_stds], axis=-1)
            list_means.append(means)
            list_stds.append(stds)

        # build list of inputs for generation model
        list_inputs = [list_label_maps, list_means, list_stds]
        if batchsize > 1:  # concatenate each input type if batchsize > 1
            list_inputs = [np.concatenate(item, 0) for item in list_inputs]
        else:
            list_inputs = [item[0] for item in list_inputs]

        yield list_inputs



"""
If you use this code, please cite one of the SynthSeg papers:
https://github.com/BBillot/SynthSeg/blob/master/bibtex.bib

Copyright 2020 Benjamin Billot

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software distributed under the License is
distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
implied. See the License for the specific language governing permissions and limitations under the
License.
"""

# python imports
import numpy as np
import tensorflow as tf
import keras.layers as KL
from keras.models import Model

# third-party imports
from lamareg.ext.lab2im import layers


def metrics_model(input_model, label_list, metrics="dice"):

    # get prediction
    last_tensor = input_model.outputs[0]
    input_shape = last_tensor.get_shape().as_list()[1:]

    # check shapes
    n_labels = input_shape[-1]
    label_list = np.unique(label_list)
    assert n_labels == len(
        label_list
    ), "label_list should be as long as the posteriors channels"

    # get GT and convert it to probabilistic values
    labels_gt = input_model.get_layer("labels_out").output
    labels_gt = layers.ConvertLabels(label_list)(labels_gt)
    labels_gt = KL.Lambda(
        lambda x: tf.one_hot(tf.cast(x, dtype="int32"), depth=n_labels, axis=-1)
    )(labels_gt)
    labels_gt = KL.Reshape(input_shape)(labels_gt)

    # make sure the tensors have the right keras shape
    last_tensor._keras_shape = tuple(last_tensor.get_shape().as_list())
    labels_gt._keras_shape = tuple(labels_gt.get_shape().as_list())

    if metrics == "dice":
        last_tensor = layers.DiceLoss()([labels_gt, last_tensor])

    elif metrics == "wl2":
        last_tensor = layers.WeightedL2Loss(target_value=5)([labels_gt, last_tensor])

    else:
        raise Exception(
            'metrics should either be "dice or "wl2, got {}'.format(metrics)
        )

    # create the model and return
    model = Model(inputs=input_model.inputs, outputs=last_tensor)
    return model


class IdentityLoss(object):
    """Very simple loss, as the computation of the loss as been directly implemented in the model."""

    def __init__(self, keepdims=True):
        self.keepdims = keepdims

    def loss(self, y_true, y_predicted):
        """Because the metrics is already calculated in the model, we simply return y_predicted.
        We still need to put y_true in the inputs, as it's expected by keras."""
        loss = y_predicted

        tf.debugging.check_numerics(loss, "Loss not finite")
        return loss



"""
Example script for contrast-agnostic registration using SynthSeg

This script demonstrates a full registration pipeline that uses SynthSeg's brain
parcellation to enable registration between images of different contrasts:

1. Generate parcellations of both input and reference images using SynthSeg
2. Register the parcellations to each other (contrast-agnostic)
3. Apply the resulting transformation to the original input image

This approach is useful for registering images with very different contrasts
(e.g., T1w to T2w, FLAIR to T1w, etc.) where direct intensity-based
registration might fail.
"""

import os
import argparse
import sys
import multiprocessing
import tempfile
import ants
from lamareg.scripts.synthseg import main as synthseg_main
from lamareg.scripts.coregister import ants_linear_nonlinear_registration
from lamareg.scripts.apply_warp import apply_warp

DEFAULT_THREADS = multiprocessing.cpu_count()

def lamareg(
    input_image,
    reference_image,
    output_image=None,
    input_parc=None,
    reference_parc=None,
    output_parc=None,
    generate_warpfield=False,
    apply_warpfield=False,
    registration_method="SyNRA",
    affine_file=None,
    warp_file=None,
    inverse_warp_file=None,
    threads=DEFAULT_THREADS,
    qc_csv=None,
    skip_fixed_parc=False,
    skip_moving_parc=False,
    skip_qc=False,
    disable_robust=False,
    inverse=False,
    secondary_warp_file=None,
    inverse_secondary_warp_file=None,
    verbose=False,
    compose=False
    
):
    """
    Perform contrast-agnostic registration using SynthSeg parcellation.
    """
    temp_files = []
    # Validate arguments based on the selected workflow
    if generate_warpfield and apply_warpfield:
        raise ValueError(
            "Cannot use both --generate-warpfield and --apply-warpfield at the same time"
        )

    # Validate input files exist
    for input_file in [f for f in [input_image, reference_image] if f is not None]:
        if not os.path.isfile(input_file):
            raise FileNotFoundError(f"Input file not found: {input_file}")

    # Validate thread counts
    if threads < 1:
        threads = DEFAULT_THREADS
    # Workflow-specific validation
    if not apply_warpfield:
        # Registration or Generate-warpfield workflow
        if input_image is None:
            raise ValueError("--moving is required for registration")
        if reference_image is None:
            raise ValueError("--fixed is required for registration")
        # For normal registration (not generate-warpfield), output image is required
        if not generate_warpfield and output_image is None:
            raise ValueError("--output is required for registration")
        
        if skip_qc:
            print("Quality control (QC) generation is disabled")

        if not disable_robust:
            print(
                "\nNote: Robust registration is enabled - a second registration stage will be performed for improved accuracy"
            )
            if compose and warp_file is not None:
                print(
                    "Note: Warp field composition is enabled - warp fields from both stages will be composed"
                )
                print(
                    "Warning: Warpfield composition can result in loss of precision and accuracy in the resulting warpfields. Consider providing a secondary warp field file path for improved results if you plan to re-use warpfields."
                )
                if secondary_warp_file is not None:
                    raise ValueError(
                        "Cannot use --compose with --secondary-warpfield. Choose one or the other."
                    )
                if inverse_secondary_warp_file is not None:
                    raise ValueError(
                        "Cannot use --compose with --inverse-secondary-warpfield. Choose one or the other."
                    )
            if not compose and warp_file is not None:
                print(
                    "Note: Warp field composition is disabled - first and second stage warp fields will be saved separately"
                )
                if secondary_warp_file is None:
                    raise ValueError(
                        "Secondary warp field file path is required when not composing warp fields. Provide --secondary-warpfield."
                    )
                if inverse_secondary_warp_file is None and inverse_warp_file is not None:
                    raise ValueError(
                        "Inverse secondary warp field file path is required when not composing warp fields. Provide --inverse-secondary-warpfield."
                    )
        else:
            print(
                "\nNote: Robust registration is disabled - only a single registration stage will be performed"
            )
            if compose:
                raise ValueError(
                    "Cannot use --compose when robust registration is disabled"
                )
            if secondary_warp_file is not None:
                raise ValueError(
                    "Cannot use --secondary-warpfield when robust registration is disabled"
                )
            if inverse_secondary_warp_file is not None:
                raise ValueError(
                    "Cannot use --inverse-secondary-warpfield when robust registration is disabled"
                )
        # If generating warpfield, warn if transform files not specified
        if affine_file is None:
            print(
                "Warning: No affine transform file path provided - affine transform will not be saved"
            )
            affine_file = tempfile.NamedTemporaryFile(suffix="_tmp_affine.mat", delete=False).name
            temp_files.append(affine_file)
        if warp_file is None:
            print(
                "Warning: No warp field file path provided - warp field will not be saved"
            )
            warp_file = tempfile.NamedTemporaryFile(suffix="_tmp_warp.nii.gz", delete=False).name
            temp_files.append(warp_file)
        if inverse_warp_file is None:
            print(
                "Warning: No inverse warp field file path provided - inverse warp field will not be saved"
            )
        if input_parc is None:
            print(
                "Warning: No moving parcellation file path provided - parcellation will not be saved"
            )
            input_parc = tempfile.NamedTemporaryFile(suffix="_tmp_moving_parc.nii.gz", delete=False).name
            temp_files.append(input_parc)
        if reference_parc is None:
            print(
                "Warning: No fixed parcellation file path provided - parcellation will not be saved"
            )
            reference_parc = tempfile.NamedTemporaryFile(suffix="_tmp_fixed_parc.nii.gz", delete=False).name
            temp_files.append(reference_parc)
        if output_parc is None:
            print(
                "Warning: No output parcellation file path provided - registered parcellation will not be saved"
            )
            output_parc = tempfile.NamedTemporaryFile(suffix="_tmp_registered_parc.nii.gz", delete=False).name
            temp_files.append(output_parc)
    else:
        # Apply-warpfield workflow
        if input_image is None:
            raise ValueError("--moving is required for apply-warpfield")
        if reference_image is None:
            raise ValueError("--fixed is required for apply-warpfield")
        if output_image is None:
            raise ValueError("--output is required for apply-warpfield")
        if affine_file is None and warp_file is None:
            raise ValueError(
                "At least one of --affine or --warpfield is required for apply-warpfield"
            )

        # Validate transform files exist
        for transform_file in [affine_file, warp_file]:
            if not os.path.isfile(transform_file):
                raise FileNotFoundError(f"Transform file not found: {transform_file}")

    # Add QC CSV validation
    if not apply_warpfield and not generate_warpfield:
        if skip_qc:
            pass
        elif qc_csv is None:
            print("Quality control (QC) skipped automatically because --qc-csv was not provided.")
            skip_qc = True
        else:
            qc_dir = os.path.dirname(qc_csv)
            if qc_dir:
                if os.path.exists(qc_dir):
                    if not os.access(qc_dir, os.W_OK):
                        raise PermissionError(
                            f"Cannot write to QC CSV directory: {qc_dir}. Check permissions."
                        )
                else:
                    try:
                        os.makedirs(qc_dir, exist_ok=True)
                    except Exception as e:
                        raise PermissionError(
                            f"Cannot create QC CSV directory: {qc_dir}. Error: {e}"
                        )

    # Create directories for all output files
    for file_path in [
        output_image,
        input_parc,
        reference_parc,
        output_parc,
        affine_file,
        warp_file,
        inverse_warp_file,
        qc_csv,
    ]:
        if file_path is not None:
            output_dir = os.path.dirname(file_path)
            if output_dir:  # Only try to create if there's a directory part
                try:
                    os.makedirs(output_dir, exist_ok=True)
                except PermissionError:
                    raise PermissionError(
                        f"Cannot create output directory: {output_dir}. Check permissions."
                    )

    print(f"Processing input image: {input_image}")
    print(f"Reference image: {reference_image}")
    print(
        f"Using {threads} thread(s)."
    )

    # Set environment variables directly
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"  # 0=ALL, 1=INFO, 2=WARNING, 3=ERROR
    os.environ["PYTHONWARNINGS"] = "ignore"
    os.environ["ITK_GLOBAL_DEFAULT_NUMBER_OF_THREADS"] = str(threads)
    os.environ["OMP_NUM_THREADS"] = str(threads)

    temp_warp_file = None
    temp_inverse_warp_file = None
    temp_parc_file = None

    try:
        # WORKFLOW 1 & 2: Full registration or generate warpfield
        if not apply_warpfield:
            # Step 1: Generate parcellations with SynthSeg if needed
            if not skip_moving_parc:
                print("\n--- Step 1.1: Generating parcellation for input image ---")
                synthseg_args = {
                    "i": input_image,
                    "o": input_parc,
                    "parc": True,
                    "robust": True,
                    "fast": False,
                    "cpu": True,
                    "threads": threads,
                    "v1": False,
                    "ct": False,
                    "vol": None,
                    "qc": None,
                    "post": None,
                    "resample": None,
                    "crop": None
                }
                synthseg_main(synthseg_args)
            else:
                print(
                    f"Skipping parcellation generation for input image: {input_image}"
                )
                print(f"Using provided parcellation: {input_parc}")
            # Check if the input parcellation file exists
            if not os.path.isfile(input_parc):
                raise FileNotFoundError(
                    f"Input parcellation file not found: {input_parc}"
                )

            if not skip_fixed_parc:
                print("\n--- Step 1.2: Generating parcellation for reference image ---")
                synthseg_args = {
                    "i": reference_image,
                    "o": reference_parc,
                    "parc": True,
                    "robust": True,
                    "fast": False,
                    "cpu": True,
                    "threads": threads,
                    "v1": False,
                    "ct": False,
                    "vol": None,
                    "qc": None,
                    "post": None,
                    "resample": None,
                    "crop": None
                }
                synthseg_main(synthseg_args)
            else:
                print(
                    f"Skipping parcellation generation for reference image: {reference_image}"
                )
                print(f"Using provided parcellation: {reference_parc}")
            # Check if the reference parcellation file exists
            if not os.path.isfile(reference_parc):
                raise FileNotFoundError(
                    f"Reference parcellation file not found: {reference_parc}"
                )

            # Step 2: Register parcellations using coregister
            print("\n--- Step 2: Coregistering parcellated images ---")
            
            # Determine output file
            current_out_file = output_parc
            if not disable_robust:
                 with tempfile.NamedTemporaryFile(suffix="_tmp_output_parc.nii.gz", delete=False) as tmp_parc:
                    temp_parc_file = tmp_parc.name
                 temp_files.append(temp_parc_file)
                 current_out_file = temp_parc_file
            
            # Determine warp file
            current_warp_file = None
            if warp_file:
                if not disable_robust:
                    if not secondary_warp_file:
                        with tempfile.NamedTemporaryFile(suffix="_tmp_warp.nii.gz", delete=False) as tmp_warp:
                            temp_warp_file = tmp_warp.name
                        temp_files.append(temp_warp_file)
                        current_warp_file = temp_warp_file
                    else:
                        current_warp_file = warp_file
                else:
                    current_warp_file = warp_file

            # Determine inverse warp file
            current_rev_warp_file = None
            if inverse_warp_file:
                if not disable_robust:
                    if not inverse_secondary_warp_file:
                        with tempfile.NamedTemporaryFile(suffix="_tmp_inverse_warp.nii.gz", delete=False) as tmp_inverse_warp:
                            temp_inverse_warp_file = tmp_inverse_warp.name
                        temp_files.append(temp_inverse_warp_file)
                        current_rev_warp_file = temp_inverse_warp_file
                    else:
                        current_rev_warp_file = inverse_warp_file
                else:
                    current_rev_warp_file = inverse_warp_file

            ants_linear_nonlinear_registration(
                fixed_file=reference_parc,
                moving_file=input_parc,
                out_file=current_out_file,
                registration_method=registration_method,
                fixed_image=reference_image,
                affine_file=affine_file,
                warp_file=current_warp_file,
                rev_warp_file=current_rev_warp_file,
                verbose=verbose,
                threads=threads
            )

            if not disable_robust:
                print(
                    "\n--- Step 2.1: Running robust registration for improved accuracy ---"
                )
                
                current_initial_warp = temp_warp_file if not secondary_warp_file else warp_file
                current_initial_inverse_warp = None
                if inverse_warp_file:
                    if not inverse_secondary_warp_file:
                        current_initial_inverse_warp = temp_inverse_warp_file
                    else:
                        current_initial_inverse_warp = inverse_warp_file
                
                current_warp_file = None
                disable_warp_comp = False
                if warp_file or not skip_qc:
                    if secondary_warp_file:
                        current_warp_file = secondary_warp_file
                        disable_warp_comp = True
                    else:
                        current_warp_file = warp_file
                
                current_rev_warp_file = None
                disable_inv_warp_comp = False
                if inverse_warp_file:
                    if inverse_secondary_warp_file:
                        current_rev_warp_file = inverse_secondary_warp_file
                        disable_inv_warp_comp = True
                    else:
                        current_rev_warp_file = inverse_warp_file

                ants_linear_nonlinear_registration(
                    fixed_file=reference_image,
                    moving_file=input_image,
                    interpolator="linear",
                    registration_method="SyNOnly",
                    initial_affine_file=affine_file,
                    initial_warp_file=current_initial_warp,
                    reg_iterations=(10, 20),
                    initial_inverse_warp_file=current_initial_inverse_warp,
                    out_file=None, # Changed from output_image to None to prevent blurred output
                    affine_file=affine_file if (affine_file or not skip_qc) else None,
                    warp_file=current_warp_file,
                    rev_warp_file=current_rev_warp_file,
                    disable_warp_composition=disable_warp_comp,
                    disable_inverse_warp_composition=disable_inv_warp_comp,
                    verbose=verbose,
                    threads=threads
                )

                try:
                    if temp_warp_file and os.path.exists(temp_warp_file):
                        os.remove(temp_warp_file)
                    if temp_inverse_warp_file and os.path.exists(temp_inverse_warp_file):
                        os.remove(temp_inverse_warp_file)
                except OSError:
                    pass

                # Apply warp to parcellation
                moving_img = ants.image_read(input_parc)
                reference_img = ants.image_read(reference_image)
                
                apply_warp(
                    moving_img=moving_img,
                    reference_img=reference_img,
                    affine_file=affine_file,
                    warp_file=warp_file,
                    out_file=output_parc,
                    interpolation="nearestNeighbor",
                    secondary_warp=secondary_warp_file,
                    verbose=verbose
                )
                    
            # Run Dice evaluation after coregistration
            if not skip_qc:


                # If qc_csv is not provided, generate a default path based on output_parc
                dice_output = (
                    qc_csv
                    if qc_csv
                    else output_parc.split('.')[0] + "_dice_scores.csv"
                )

                print(
                    "\n--- Step 3: Calculating Dice scores to evaluate registration quality ---"
                )
                try:
                    from lamareg.scripts.dice_compare import compare_parcellations_dice

                    compare_parcellations_dice(reference_parc, output_parc, dice_output, verbose=verbose)
                    print(f"Quality control metrics saved to: {dice_output}")
                except FileNotFoundError as e:
                    print(
                        f"Warning: Could not calculate Dice scores - file not found: {e}",
                        file=sys.stderr,
                    )
                except PermissionError as e:
                    print(
                        f"Warning: Could not calculate Dice scores - permission error: {e}",
                        file=sys.stderr,
                    )
                except ImportError as e:
                    print(
                        f"Warning: Could not calculate Dice scores - dice_compare module not found",
                        file=sys.stderr,
                    )
                except Exception as e:
                    print(
                        f"Warning: Could not calculate Dice scores: {e}",
                        file=sys.stderr,
                    )

        # WORKFLOW 1 & 3: Apply transformation to the original input image
        # Run this step for ALL registration cases if an output image is requested
        # Previously this was restricted: (disable_robust or apply_warpfield)
        if not generate_warpfield and output_image is not None:
            print(
                "\n--- Step 4: Applying transformation to original input image ---"
            )
            
            moving_img = ants.image_read(input_image)
            reference_img = ants.image_read(reference_image)

            apply_warp(
                moving_img=moving_img,
                reference_img=reference_img,
                affine_file=affine_file,
                warp_file=warp_file,
                out_file=output_image,
                secondary_warp=secondary_warp_file,
                inverse=inverse,
                verbose=verbose
            )

            print(f"\nSuccess! Registered image saved to: {output_image}")
        elif generate_warpfield:
            success_msg = "\nSuccess! "
            if warp_file:
                success_msg += f"Warp field generated at: {warp_file}"
            if affine_file:
                success_msg += f"\nAffine transformation saved at: {affine_file}"
            print(success_msg)

    except Exception as e:
        print(f"Error during processing: {e}", file=sys.stderr)
        sys.exit(1)
    finally:
        for temp_path in temp_files:
            if temp_path and os.path.exists(temp_path):
                try:
                    os.remove(temp_path)
                except OSError:
                    pass


def main():
    """Entry point for command-line use"""
    parser = argparse.ArgumentParser(
        description="Contrast-agnostic registration using SynthSeg"
    )
    parser.add_argument(
        "--moving", required=True, help="Input moving image to be registered"
    )
    parser.add_argument(
        "--fixed", required=True, help="Reference fixed image (target space)"
    )
    parser.add_argument(
        "--skip-fixed-parc",
        action="store_true",
        help="Skip generating fixed parcellation",
    )
    parser.add_argument(
        "--skip-moving-parc",
        action="store_true",
        help="Skip generating moving parcellation",
    )
    parser.add_argument("--output", required=True, help="Output registered image")
    parser.add_argument(
        "--moving-parc", help="Path for moving image parcellation"
    )
    parser.add_argument(
        "--fixed-parc", help="Path for fixed image parcellation"
    )
    parser.add_argument(
        "--registered-parc", help="Path for registered parcellation"
    )
    parser.add_argument(
        "--affine", help="Path for affine transformation"
    )
    parser.add_argument("--warpfield", help="Path for warp field")
    parser.add_argument("--inverse-warpfield", help="Path for inverse warp field")
    parser.add_argument(
        "--generate-warpfield",
        action="store_true",
        help="Generate warp field without applying it",
    )
    parser.add_argument(
        "--apply-warpfield",
        action="store_true",
        help="Apply existing warp field to moving image",
    )
    parser.add_argument(
        "--registration-method", default="SyNRA", help="Registration method"
    )
    parser.add_argument(
        "--threads",
        type=int,
        default=DEFAULT_THREADS,
        help="Number of threads to use for processing",
    )

    parser.add_argument("--secondary-warpfield", help="Path for secondary warp field (for robust registration)")
    parser.add_argument("--inverse-secondary-warpfield", help="Path for inverse of secondary warp field (for robust registration)")
    parser.add_argument("--qc-csv", help="Path for quality control Dice score CSV file")
    parser.add_argument("--skip-qc", action="store_true", help="Skip QC CSV generation")
    parser.add_argument("--disable-robust", action="store_true", help="Disable robust second-stage registration")
    parser.add_argument("--inverse", action="store_true", help="Whether to reverse the order of the transforms (warpfield first, then affine)")
    parser.add_argument("--verbose", action="store_true", help="Enable verbose output")
    parser.add_argument("--compose", action="store_true", help="Compose warp fields instead of applying sequentially")
    args = parser.parse_args()

    # Validate arguments based on workflow
    if args.apply_warpfield and (args.affine is None or args.warpfield is None):
        parser.error("--apply-warpfield requires --affine and --warpfield arguments")

    if args.generate_warpfield and args.output is not None:
        parser.error(
            "--generate-warpfield cannot be used with --output (no output image is produced)"
        )

    if not args.apply_warpfield and not args.generate_warpfield and args.output is None:
        parser.error(
            "--output is required unless --generate-warpfield or --apply-warpfield is specified"
        )

    lamareg(
        input_image=args.moving,
        reference_image=args.fixed,
        output_image=args.output,
        input_parc=args.moving_parc,
        reference_parc=args.fixed_parc,
        output_parc=args.registered_parc,
        generate_warpfield=args.generate_warpfield,
        apply_warpfield=args.apply_warpfield,
        registration_method=args.registration_method,
        affine_file=args.affine,
        warp_file=args.warpfield,
        inverse_warp_file=args.inverse_warpfield,
        threads=args.threads,
        qc_csv=args.qc_csv,
        skip_fixed_parc=args.skip_fixed_parc,
        skip_moving_parc=args.skip_moving_parc,
        skip_qc=args.skip_qc,
        disable_robust=args.disable_robust,
        inverse=args.inverse,
        secondary_warp_file=args.secondary_warpfield,
        inverse_secondary_warp_file=args.inverse_secondary_warpfield,
        verbose=args.verbose,
        compose=args.compose
    )


if __name__ == "__main__":
    main()



"""LaMAR scripts package."""

# Import main functions to make them available at package level
from lamareg.scripts.lamar import lamareg
from lamareg.scripts import synthseg, coregister, apply_warp

__all__ = [
    'lamareg',
    'synthseg', 
    'coregister',
    'apply_warp',
]


"""
If you use this code, please cite the first SynthSeg paper:
https://github.com/BBillot/lab2im/blob/master/bibtex.bib

Copyright 2020 Benjamin Billot

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software distributed under the License is
distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
implied. See the License for the specific language governing permissions and limitations under the
License.
"""

# python imports
import numpy as np
import keras.layers as KL
from keras.models import Model

# project imports
from lamareg.ext.lab2im import utils
from lamareg.ext.lab2im import layers
from lamareg.ext.lab2im.edit_tensors import (
    resample_tensor,
    blurring_sigma_for_downsampling,
)


def lab2im_model(
    labels_shape,
    n_channels,
    generation_labels,
    output_labels,
    atlas_res,
    target_res,
    output_shape=None,
    output_div_by_n=None,
    blur_range=1.15,
):
    """
    This function builds a keras/tensorflow model to generate images from provided label maps.
    The images are generated by sampling a Gaussian Mixture Model (of given parameters), conditioned on the label map.
    The model will take as inputs:
        -a label map
        -a vector containing the means of the Gaussian Mixture Model for each label,
        -a vector containing the standard deviations of the Gaussian Mixture Model for each label,
        -an array of size batch*(n_dims+1)*(n_dims+1) representing a linear transformation
    The model returns:
        -the generated image normalised between 0 and 1.
        -the corresponding label map, with only the labels present in output_labels (the other are reset to zero).
    :param labels_shape: shape of the input label maps. Can be a sequence or a 1d numpy array.
    :param n_channels: number of channels to be synthetised.
    :param generation_labels: list of all possible label values in the input label maps.
    Can be a sequence or a 1d numpy array.
    :param output_labels: list of the same length as generation_labels to indicate which values to use in the label maps
    returned by this model, i.e. all occurrences of generation_labels[i] in the input label maps will be converted to
    output_labels[i] in the returned label maps. Examples:
    Set output_labels[i] to zero if you wish to erase the value generation_labels[i] from the returned label maps.
    Set output_labels[i]=generation_labels[i] if you wish to keep the value generation_labels[i] in the returned maps.
    Can be a list or a 1d numpy array. By default output_labels is equal to generation_labels.
    :param atlas_res: resolution of the input label maps.
    Can be a number (isotropic resolution), a sequence, or a 1d numpy array.
    :param target_res: target resolution of the generated images and corresponding label maps.
    Can be a number (isotropic resolution), a sequence, or a 1d numpy array.
    :param output_shape: (optional) desired shape of the output images.
    If the atlas and target resolutions are the same, the output will be cropped to output_shape, and if the two
    resolutions are different, the output will be resized with trilinear interpolation to output_shape.
    Can be an integer (same size in all dimensions), a sequence, or a 1d numpy array.
    :param output_div_by_n: (optional) forces the output shape to be divisible by this value. It overwrites output_shape
    if necessary. Can be an integer (same size in all dimensions), a sequence, or a 1d numpy array.
    :param blur_range: (optional) Randomise the standard deviation of the blurring kernels, (whether data_res is given
    or not). At each mini_batch, the standard deviation of the blurring kernels are multiplied by a coefficient sampled
    from a uniform distribution with bounds [1/blur_range, blur_range]. If None, no randomisation. Default is 1.15.
    """

    # reformat resolutions
    labels_shape = utils.reformat_to_list(labels_shape)
    n_dims, _ = utils.get_dims(labels_shape)
    atlas_res = utils.reformat_to_n_channels_array(atlas_res, n_dims=n_dims)[0]
    target_res = (
        atlas_res
        if (target_res is None)
        else utils.reformat_to_n_channels_array(target_res, n_dims)[0]
    )

    # get shapes
    crop_shape, output_shape = get_shapes(
        labels_shape, output_shape, atlas_res, target_res, output_div_by_n
    )

    # define model inputs
    labels_input = KL.Input(
        shape=labels_shape + [1], name="labels_input", dtype="int32"
    )
    means_input = KL.Input(
        shape=list(generation_labels.shape) + [n_channels], name="means_input"
    )
    stds_input = KL.Input(
        shape=list(generation_labels.shape) + [n_channels], name="stds_input"
    )

    # deform labels
    labels = layers.RandomSpatialDeformation(inter_method="nearest")(labels_input)

    # cropping
    if crop_shape != labels_shape:
        labels._keras_shape = tuple(labels.get_shape().as_list())
        labels = layers.RandomCrop(crop_shape)(labels)

    # build synthetic image
    labels._keras_shape = tuple(labels.get_shape().as_list())
    image = layers.SampleConditionalGMM(generation_labels)(
        [labels, means_input, stds_input]
    )

    # apply bias field
    image._keras_shape = tuple(image.get_shape().as_list())
    image = layers.BiasFieldCorruption(0.3, 0.025, same_bias_for_all_channels=False)(
        image
    )

    # intensity augmentation
    image._keras_shape = tuple(image.get_shape().as_list())
    image = layers.IntensityAugmentation(clip=300, normalise=True, gamma_std=0.2)(image)

    # blur image
    sigma = blurring_sigma_for_downsampling(atlas_res, target_res)
    image._keras_shape = tuple(image.get_shape().as_list())
    image = layers.GaussianBlur(sigma=sigma, random_blur_range=blur_range)(image)

    # resample to target res
    if crop_shape != output_shape:
        image = resample_tensor(image, output_shape, interp_method="linear")
        labels = resample_tensor(labels, output_shape, interp_method="nearest")

    # reset unwanted labels to zero
    labels = layers.ConvertLabels(
        generation_labels, dest_values=output_labels, name="labels_out"
    )(labels)

    # build model (dummy layer enables to keep the labels when plugging this model to other models)
    image = KL.Lambda(lambda x: x[0], name="image_out")([image, labels])
    brain_model = Model(
        inputs=[labels_input, means_input, stds_input], outputs=[image, labels]
    )

    return brain_model


def get_shapes(labels_shape, output_shape, atlas_res, target_res, output_div_by_n):

    n_dims = len(atlas_res)

    # get resampling factor
    if atlas_res.tolist() != target_res.tolist():
        resample_factor = [atlas_res[i] / float(target_res[i]) for i in range(n_dims)]
    else:
        resample_factor = None

    # output shape specified, need to get cropping shape, and resample shape if necessary
    if output_shape is not None:
        output_shape = utils.reformat_to_list(output_shape, length=n_dims, dtype="int")

        # make sure that output shape is smaller or equal to label shape
        if resample_factor is not None:
            output_shape = [
                min(int(labels_shape[i] * resample_factor[i]), output_shape[i])
                for i in range(n_dims)
            ]
        else:
            output_shape = [
                min(labels_shape[i], output_shape[i]) for i in range(n_dims)
            ]

        # make sure output shape is divisible by output_div_by_n
        if output_div_by_n is not None:
            tmp_shape = [
                utils.find_closest_number_divisible_by_m(s, output_div_by_n)
                for s in output_shape
            ]
            if output_shape != tmp_shape:
                print(
                    "output shape {0} not divisible by {1}, changed to {2}".format(
                        output_shape, output_div_by_n, tmp_shape
                    )
                )
                output_shape = tmp_shape

        # get cropping and resample shape
        if resample_factor is not None:
            cropping_shape = [
                int(np.around(output_shape[i] / resample_factor[i], 0))
                for i in range(n_dims)
            ]
        else:
            cropping_shape = output_shape

    # no output shape specified, so no cropping unless label_shape is not divisible by output_div_by_n
    else:
        cropping_shape = labels_shape
        if resample_factor is not None:
            output_shape = [
                int(np.around(cropping_shape[i] * resample_factor[i], 0))
                for i in range(n_dims)
            ]
        else:
            output_shape = cropping_shape
        # make sure output shape is divisible by output_div_by_n
        if output_div_by_n is not None:
            output_shape = [
                utils.find_closest_number_divisible_by_m(
                    s, output_div_by_n, answer_type="closer"
                )
                for s in output_shape
            ]

    return cropping_shape, output_shape



"""
If you use this code, please cite one of the SynthSeg papers:
https://github.com/BBillot/SynthSeg/blob/master/bibtex.bib

Copyright 2020 Benjamin Billot

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software distributed under the License is
distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
implied. See the License for the specific language governing permissions and limitations under the
License.
"""

# python imports
import os
import numpy as np
from scipy.stats import wilcoxon
from scipy.ndimage.morphology import distance_transform_edt

# third-party imports
from lamareg.ext.lab2im import utils
from lamareg.ext.lab2im import edit_volumes


def fast_dice(x, y, labels):
    """Fast implementation of Dice scores.
    :param x: input label map
    :param y: input label map of the same size as x
    :param labels: numpy array of labels to evaluate on
    :return: numpy array with Dice scores in the same order as labels.
    """

    assert (
        x.shape == y.shape
    ), "both inputs should have same size, had {} and {}".format(x.shape, y.shape)

    if len(labels) > 1:
        # sort labels
        labels_sorted = np.sort(labels)

        # build bins for histograms
        label_edges = np.sort(
            np.concatenate([labels_sorted - 0.1, labels_sorted + 0.1])
        )
        label_edges = np.insert(
            label_edges,
            [0, len(label_edges)],
            [labels_sorted[0] - 0.1, labels_sorted[-1] + 0.1],
        )

        # compute Dice and re-arrange scores in initial order
        hst = np.histogram2d(x.flatten(), y.flatten(), bins=label_edges)[0]
        idx = np.arange(start=1, stop=2 * len(labels_sorted), step=2)
        dice_score = (
            2 * np.diag(hst)[idx] / (np.sum(hst, 0)[idx] + np.sum(hst, 1)[idx] + 1e-5)
        )
        dice_score = dice_score[np.searchsorted(labels_sorted, labels)]

    else:
        dice_score = dice(x == labels[0], y == labels[0])

    return dice_score


def dice(x, y):
    """Implementation of dice scores for 0/1 numpy array"""
    return 2 * np.sum(x * y) / (np.sum(x) + np.sum(y))


def surface_distances(
    x, y, hausdorff_percentile=None, return_coordinate_max_distance=False
):
    """Computes the maximum boundary distance (Hausdorff distance), and the average boundary distance of two masks.
    :param x: numpy array (boolean or 0/1)
    :param y: numpy array (boolean or 0/1)
    :param hausdorff_percentile: (optional) percentile (from 0 to 100) for which to compute the Hausdorff distance.
    Set this to 100 to compute the real Hausdorff distance (default). Can also be a list, where HD will be computed for
    the provided values.
    :param return_coordinate_max_distance: (optional) when set to true, the function will return the coordinates of the
    voxel with the highest distance (only if hausdorff_percentile=100).
    :return: max_dist, mean_dist(, coordinate_max_distance)
    max_dist: scalar with HD computed for the given percentile (or list if hausdorff_percentile was given as a list).
    mean_dist: scalar with average surface distance
    coordinate_max_distance: only returned return_coordinate_max_distance is True."""

    assert (
        x.shape == y.shape
    ), "both inputs should have same size, had {} and {}".format(x.shape, y.shape)
    n_dims = len(x.shape)

    hausdorff_percentile = 100 if hausdorff_percentile is None else hausdorff_percentile
    hausdorff_percentile = utils.reformat_to_list(hausdorff_percentile)

    # crop x and y around ROI
    _, crop_x = edit_volumes.crop_volume_around_region(x)
    _, crop_y = edit_volumes.crop_volume_around_region(y)

    # set distances to maximum volume shape if they are not defined
    if (crop_x is None) | (crop_y is None):
        return max(x.shape), max(x.shape)

    crop = np.concatenate(
        [np.minimum(crop_x, crop_y)[:n_dims], np.maximum(crop_x, crop_y)[n_dims:]]
    )
    x = edit_volumes.crop_volume_with_idx(x, crop)
    y = edit_volumes.crop_volume_with_idx(y, crop)

    # detect edge
    x_dist_int = distance_transform_edt(x * 1)
    x_edge = (x_dist_int == 1) * 1
    y_dist_int = distance_transform_edt(y * 1)
    y_edge = (y_dist_int == 1) * 1

    # calculate distance from edge
    x_dist = distance_transform_edt(np.logical_not(x_edge))
    y_dist = distance_transform_edt(np.logical_not(y_edge))

    # find distances from the 2 surfaces
    x_dists_to_y = y_dist[x_edge == 1]
    y_dists_to_x = x_dist[y_edge == 1]

    max_dist = list()
    coordinate_max_distance = None
    for hd_percentile in hausdorff_percentile:

        # find max distance from the 2 surfaces
        if hd_percentile == 100:
            max_dist.append(np.max(np.concatenate([x_dists_to_y, y_dists_to_x])))

            if return_coordinate_max_distance:
                indices_x_surface = np.where(x_edge == 1)
                idx_max_distance_x = np.where(x_dists_to_y == max_dist)[0]
                if idx_max_distance_x.size != 0:
                    coordinate_max_distance = np.stack(indices_x_surface).transpose()[
                        idx_max_distance_x
                    ]
                else:
                    indices_y_surface = np.where(y_edge == 1)
                    idx_max_distance_y = np.where(y_dists_to_x == max_dist)[0]
                    coordinate_max_distance = np.stack(indices_y_surface).transpose()[
                        idx_max_distance_y
                    ]

        # find percentile of max distance
        else:
            max_dist.append(
                np.percentile(
                    np.concatenate([x_dists_to_y, y_dists_to_x]), hd_percentile
                )
            )

    # find average distance between 2 surfaces
    if x_dists_to_y.shape[0] > 0:
        x_mean_dist_to_y = np.mean(x_dists_to_y)
    else:
        x_mean_dist_to_y = max(x.shape)
    if y_dists_to_x.shape[0] > 0:
        y_mean_dist_to_x = np.mean(y_dists_to_x)
    else:
        y_mean_dist_to_x = max(x.shape)
    mean_dist = (x_mean_dist_to_y + y_mean_dist_to_x) / 2

    # convert max dist back to scalar if HD only computed for 1 percentile
    if len(max_dist) == 1:
        max_dist = max_dist[0]

    # return coordinate of max distance if necessary
    if coordinate_max_distance is not None:
        return max_dist, mean_dist, coordinate_max_distance
    else:
        return max_dist, mean_dist


def compute_non_parametric_paired_test(
    dice_ref, dice_compare, eval_indices=None, alternative="two-sided"
):
    """Compute non-parametric paired t-tests between two sets of Dice scores.
    :param dice_ref: numpy array with Dice scores, rows represent structures, and columns represent subjects.
    Taken as reference for one-sided tests.
    :param dice_compare: numpy array of the same format as dice_ref.
    :param eval_indices: (optional) list or 1d array indicating the row indices of structures to run the tests for.
    Default is None, for which p-values are computed for all rows.
    :param alternative: (optional) The alternative hypothesis to be tested, can be 'two-sided', 'greater', 'less'.
    :return: 1d numpy array, with p-values for all tests on evaluated structures, as well as an additional test for
    average scores (last value of the array). The average score is computed only on the evaluation structures.
    """

    # take all rows if indices not specified
    if eval_indices is None:
        if len(dice_ref.shape) > 1:
            eval_indices = np.arange(dice_ref.shape[0])
        else:
            eval_indices = []

    # loop over all evaluation label values
    pvalues = list()
    if len(eval_indices) > 1:
        for idx in eval_indices:

            x = dice_ref[idx, :]
            y = dice_compare[idx, :]
            _, p = wilcoxon(x, y, alternative=alternative)
            pvalues.append(p)

        # average score
        x = np.mean(dice_ref[eval_indices, :], axis=0)
        y = np.mean(dice_compare[eval_indices, :], axis=0)
        _, p = wilcoxon(x, y, alternative=alternative)
        pvalues.append(p)

    else:
        # average score
        _, p = wilcoxon(dice_ref, dice_compare, alternative=alternative)
        pvalues.append(p)

    return np.array(pvalues)


def cohens_d(volumes_x, volumes_y):

    means_x = np.mean(volumes_x, axis=0)
    means_y = np.mean(volumes_y, axis=0)

    var_x = np.var(volumes_x, axis=0)
    var_y = np.var(volumes_y, axis=0)

    n_x = np.shape(volumes_x)[0]
    n_y = np.shape(volumes_y)[0]

    std = np.sqrt(((n_x - 1) * var_x + (n_y - 1) * var_y) / (n_x + n_y - 2))
    cohensd = (means_x - means_y) / std

    return cohensd


def evaluation(
    gt_dir,
    seg_dir,
    label_list,
    mask_dir=None,
    compute_score_whole_structure=False,
    path_dice=None,
    path_hausdorff=None,
    path_hausdorff_99=None,
    path_hausdorff_95=None,
    path_mean_distance=None,
    crop_margin_around_gt=10,
    list_incorrect_labels=None,
    list_correct_labels=None,
    use_nearest_label=False,
    recompute=True,
    verbose=True,
):
    """This function computes Dice scores, as well as surface distances, between two sets of labels maps in gt_dir
    (ground truth) and seg_dir (typically predictions). Label maps in both folders are matched by sorting order.
    The resulting scores are saved at the specified locations.
    :param gt_dir: path of directory with gt label maps
    :param seg_dir: path of directory with label maps to compare to gt_dir. Matched to gt label maps by sorting order.
    :param label_list: list of label values for which to compute evaluation metrics. Can be a sequence, a 1d numpy
    array, or the path to such array.
    :param mask_dir: (optional) path of directory with masks of areas to ignore for each evaluated segmentation.
    Matched to gt label maps by sorting order. Default is None, where nothing is masked.
    :param compute_score_whole_structure: (optional) whether to also compute the selected scores for the whole segmented
    structure (i.e. scores are computed for a single structure obtained by regrouping all non-zero values). If True, the
    resulting scores are added as an extra row to the result matrices. Default is False.
    :param path_dice: path where the resulting Dice will be writen as numpy array.
    Default is None, where the array is not saved.
    :param path_hausdorff: path where the resulting Hausdorff distances will be writen as numpy array (only if
    compute_distances is True). Default is None, where the array is not saved.
    :param path_hausdorff_99: same as for path_hausdorff but for the 99th percentile of the boundary distance.
    :param path_hausdorff_95: same as for path_hausdorff but for the 95th percentile of the boundary distance.
    :param path_mean_distance: path where the resulting mean distances will be writen as numpy array (only if
    compute_distances is True). Default is None, where the array is not saved.
    :param crop_margin_around_gt: (optional) margin by which to crop around the gt volumes, in order to compute the
    scores more efficiently. If 0, no cropping is performed.
    :param list_incorrect_labels: (optional) this option enables to replace some label values in the maps in seg_dir by
    other label values. Can be a list, a 1d numpy array, or the path to such an array.
    The incorrect labels can then be replaced either by specified values, or by the nearest value (see below).
    :param list_correct_labels: (optional) list of values to correct the labels specified in list_incorrect_labels.
    Correct values must have the same order as their corresponding value in list_incorrect_labels.
    :param use_nearest_label: (optional) whether to correct the incorrect label values with the nearest labels.
    :param recompute: (optional) whether to recompute the already existing results. Default is True.
    :param verbose: (optional) whether to print out info about the remaining number of cases.
    """

    # check whether to recompute
    compute_dice = not os.path.isfile(path_dice) if (path_dice is not None) else True
    compute_hausdorff = (
        not os.path.isfile(path_hausdorff) if (path_hausdorff is not None) else False
    )
    compute_hausdorff_99 = (
        not os.path.isfile(path_hausdorff_99)
        if (path_hausdorff_99 is not None)
        else False
    )
    compute_hausdorff_95 = (
        not os.path.isfile(path_hausdorff_95)
        if (path_hausdorff_95 is not None)
        else False
    )
    compute_mean_dist = (
        not os.path.isfile(path_mean_distance)
        if (path_mean_distance is not None)
        else False
    )
    compute_hd = [compute_hausdorff, compute_hausdorff_99, compute_hausdorff_95]

    if compute_dice | any(compute_hd) | compute_mean_dist | recompute:

        # get list label maps to compare
        path_gt_labels = utils.list_images_in_folder(gt_dir)
        path_segs = utils.list_images_in_folder(seg_dir)
        path_gt_labels = utils.reformat_to_list(path_gt_labels, length=len(path_segs))
        if len(path_gt_labels) != len(path_segs):
            print(
                "gt and segmentation folders must have the same amount of label maps."
            )
        if mask_dir is not None:
            path_masks = utils.list_images_in_folder(mask_dir)
            if len(path_masks) != len(path_segs):
                print("not the same amount of masks and segmentations.")
        else:
            path_masks = [None] * len(path_segs)

        # load labels list
        label_list, _ = utils.get_list_labels(label_list=label_list, labels_dir=gt_dir)
        n_labels = len(label_list)
        max_label = np.max(label_list) + 1

        # initialise result matrices
        if compute_score_whole_structure:
            max_dists = np.zeros((n_labels + 1, len(path_segs), 3))
            mean_dists = np.zeros((n_labels + 1, len(path_segs)))
            dice_coefs = np.zeros((n_labels + 1, len(path_segs)))
        else:
            max_dists = np.zeros((n_labels, len(path_segs), 3))
            mean_dists = np.zeros((n_labels, len(path_segs)))
            dice_coefs = np.zeros((n_labels, len(path_segs)))

        # loop over segmentations
        loop_info = utils.LoopInfo(len(path_segs), 10, "evaluating", print_time=True)
        for idx, (path_gt, path_seg, path_mask) in enumerate(
            zip(path_gt_labels, path_segs, path_masks)
        ):
            if verbose:
                loop_info.update(idx)

            # load gt labels and segmentation
            gt_labels = utils.load_volume(path_gt, dtype="int", aff_ref=np.eye(4))
            seg = utils.load_volume(path_seg, dtype="int", aff_ref=np.eye(4))
            if path_mask is not None:
                mask = utils.load_volume(path_mask, dtype="bool", aff_ref=np.eye(4))
                gt_labels[mask] = max_label
                seg[mask] = max_label

            # crop images
            if crop_margin_around_gt > 0:
                gt_labels, cropping = edit_volumes.crop_volume_around_region(
                    gt_labels, margin=crop_margin_around_gt
                )
                seg = edit_volumes.crop_volume_with_idx(seg, cropping)

            if list_incorrect_labels is not None:
                seg = edit_volumes.correct_label_map(
                    seg, list_incorrect_labels, list_correct_labels, use_nearest_label
                )

            # compute Dice scores
            dice_coefs[:n_labels, idx] = fast_dice(gt_labels, seg, label_list)

            # compute Dice scores for whole structures
            if compute_score_whole_structure:
                temp_gt = (gt_labels > 0) * 1
                temp_seg = (seg > 0) * 1
                dice_coefs[-1, idx] = dice(temp_gt, temp_seg)
            else:
                temp_gt = temp_seg = None

            # compute average and Hausdorff distances
            if any(compute_hd) | compute_mean_dist:

                # compute unique label values
                unique_gt_labels = np.unique(gt_labels)
                unique_seg_labels = np.unique(seg)

                # compute max/mean surface distances for all labels
                for index, label in enumerate(label_list):
                    if (label in unique_gt_labels) & (label in unique_seg_labels):
                        mask_gt = np.where(gt_labels == label, True, False)
                        mask_seg = np.where(seg == label, True, False)
                        tmp_max_dists, mean_dists[index, idx] = surface_distances(
                            mask_gt, mask_seg, [100, 99, 95]
                        )
                        max_dists[index, idx, :] = np.array(tmp_max_dists)
                    else:
                        mean_dists[index, idx] = max(gt_labels.shape)
                        max_dists[index, idx, :] = np.array([max(gt_labels.shape)] * 3)

                # compute max/mean distances for whole structure
                if compute_score_whole_structure:
                    tmp_max_dists, mean_dists[-1, idx] = surface_distances(
                        temp_gt, temp_seg, [100, 99, 95]
                    )
                    max_dists[-1, idx, :] = np.array(tmp_max_dists)

        # write results
        if path_dice is not None:
            utils.mkdir(os.path.dirname(path_dice))
            np.save(path_dice, dice_coefs)
        if path_hausdorff is not None:
            utils.mkdir(os.path.dirname(path_hausdorff))
            np.save(path_hausdorff, max_dists[..., 0])
        if path_hausdorff_99 is not None:
            utils.mkdir(os.path.dirname(path_hausdorff_99))
            np.save(path_hausdorff_99, max_dists[..., 1])
        if path_hausdorff_95 is not None:
            utils.mkdir(os.path.dirname(path_hausdorff_95))
            np.save(path_hausdorff_95, max_dists[..., 2])
        if path_mean_distance is not None:
            utils.mkdir(os.path.dirname(path_mean_distance))
            np.save(path_mean_distance, mean_dists)



"""

This file contains functions to handle keras/tensorflow tensors.
    - blurring_sigma_for_downsampling
    - gaussian_kernel
    - resample_tensor
    - expand_dims


If you use this code, please cite the first SynthSeg paper:
https://github.com/BBillot/lab2im/blob/master/bibtex.bib

Copyright 2020 Benjamin Billot

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software distributed under the License is
distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
implied. See the License for the specific language governing permissions and limitations under the
License.

"""

# python imports
import numpy as np
import tensorflow as tf
import keras.layers as KL
import keras.backend as K
from itertools import combinations

# project imports
from lamareg.ext.lab2im import utils

# third-party imports
import lamareg.ext.neuron.layers as nrn_layers
from lamareg.ext.neuron.utils import volshape_to_meshgrid


def blurring_sigma_for_downsampling(
    current_res, downsample_res, mult_coef=None, thickness=None
):
    """Compute standard deviations of 1d gaussian masks for image blurring before downsampling.
    :param downsample_res: resolution to downsample to. Can be a 1d numpy array or list, or a tensor.
    :param current_res: resolution of the volume before downsampling.
    Can be a 1d numpy array or list or tensor of the same length as downsample res.
    :param mult_coef: (optional) multiplicative coefficient for the blurring kernel. Default is 0.75.
    :param thickness: (optional) slice thickness in each dimension. Must be the same type as downsample_res.
    :return: standard deviation of the blurring masks given as the same type as downsample_res (list or tensor).
    """

    if not tf.is_tensor(downsample_res):

        # get blurring resolution (min between downsample_res and thickness)
        current_res = np.array(current_res)
        downsample_res = np.array(downsample_res)
        if thickness is not None:
            downsample_res = np.minimum(downsample_res, np.array(thickness))

        # get std deviation for blurring kernels
        if mult_coef is None:
            sigma = 0.75 * downsample_res / current_res
            sigma[downsample_res == current_res] = 0.5
        else:
            sigma = mult_coef * downsample_res / current_res
        sigma[downsample_res == 0] = 0

    else:

        # reformat data resolution at which we blur
        if thickness is not None:
            down_res = KL.Lambda(lambda x: tf.math.minimum(x[0], x[1]))(
                [downsample_res, thickness]
            )
        else:
            down_res = downsample_res

        # get std deviation for blurring kernels
        if mult_coef is None:
            sigma = KL.Lambda(
                lambda x: tf.where(
                    tf.math.equal(
                        x, tf.convert_to_tensor(current_res, dtype="float32")
                    ),
                    0.5,
                    0.75 * x / tf.convert_to_tensor(current_res, dtype="float32"),
                )
            )(down_res)
        else:
            sigma = KL.Lambda(
                lambda x: mult_coef
                * x
                / tf.convert_to_tensor(current_res, dtype="float32")
            )(down_res)
        sigma = KL.Lambda(lambda x: tf.where(tf.math.equal(x[0], 0.0), 0.0, x[1]))(
            [down_res, sigma]
        )

    return sigma


def gaussian_kernel(sigma, max_sigma=None, blur_range=None, separable=True):
    """Build gaussian kernels of the specified standard deviation. The outputs are given as tensorflow tensors.
    :param sigma: standard deviation of the tensors. Can be given as a list/numpy array or as tensors. In each case,
    sigma must have the same length as the number of dimensions of the volume that will be blurred with the output
    tensors (e.g. sigma must have 3 values for 3D volumes).
    :param max_sigma:
    :param blur_range:
    :param separable:
    :return:
    """
    # convert sigma into a tensor
    if not tf.is_tensor(sigma):
        sigma_tens = tf.convert_to_tensor(
            utils.reformat_to_list(sigma), dtype="float32"
        )
    else:
        assert (
            max_sigma is not None
        ), "max_sigma must be provided when sigma is given as a tensor"
        sigma_tens = sigma
    shape = sigma_tens.get_shape().as_list()

    # get n_dims and batchsize
    if shape[0] is not None:
        n_dims = shape[0]
        batchsize = None
    else:
        n_dims = shape[1]
        batchsize = tf.split(tf.shape(sigma_tens), [1, -1])[0]

    # reformat max_sigma
    if max_sigma is not None:  # dynamic blurring
        max_sigma = np.array(utils.reformat_to_list(max_sigma, length=n_dims))
    else:  # sigma is fixed
        max_sigma = np.array(utils.reformat_to_list(sigma, length=n_dims))

    # randomise the burring std dev and/or split it between dimensions
    if blur_range is not None:
        if blur_range != 1:
            sigma_tens = sigma_tens * tf.random.uniform(
                tf.shape(sigma_tens), minval=1 / blur_range, maxval=blur_range
            )

    # get size of blurring kernels
    windowsize = np.int32(np.ceil(2.5 * max_sigma) / 2) * 2 + 1

    if separable:

        split_sigma = tf.split(sigma_tens, [1] * n_dims, axis=-1)

        kernels = list()
        comb = np.array(list(combinations(list(range(n_dims)), n_dims - 1))[::-1])
        for i, wsize in enumerate(windowsize):

            if wsize > 1:

                # build meshgrid and replicate it along batch dim if dynamic blurring
                locations = tf.cast(tf.range(0, wsize), "float32") - (wsize - 1) / 2
                if batchsize is not None:
                    locations = tf.tile(
                        tf.expand_dims(locations, axis=0),
                        tf.concat(
                            [
                                batchsize,
                                tf.ones(tf.shape(tf.shape(locations)), dtype="int32"),
                            ],
                            axis=0,
                        ),
                    )
                    comb[i] += 1

                # compute gaussians
                exp_term = -K.square(locations) / (2 * split_sigma[i] ** 2)
                g = tf.exp(exp_term - tf.math.log(np.sqrt(2 * np.pi) * split_sigma[i]))
                g = g / tf.reduce_sum(g)

                for axis in comb[i]:
                    g = tf.expand_dims(g, axis=axis)
                kernels.append(tf.expand_dims(tf.expand_dims(g, -1), -1))

            else:
                kernels.append(None)

    else:

        # build meshgrid
        mesh = [
            tf.cast(f, "float32")
            for f in volshape_to_meshgrid(windowsize, indexing="ij")
        ]
        diff = tf.stack(
            [mesh[f] - (windowsize[f] - 1) / 2 for f in range(len(windowsize))], axis=-1
        )

        # replicate meshgrid to batch size and reshape sigma_tens
        if batchsize is not None:
            diff = tf.tile(
                tf.expand_dims(diff, axis=0),
                tf.concat(
                    [batchsize, tf.ones(tf.shape(tf.shape(diff)), dtype="int32")],
                    axis=0,
                ),
            )
            for i in range(n_dims):
                sigma_tens = tf.expand_dims(sigma_tens, axis=1)
        else:
            for i in range(n_dims):
                sigma_tens = tf.expand_dims(sigma_tens, axis=0)

        # compute gaussians
        sigma_is_0 = tf.equal(sigma_tens, 0)
        exp_term = -tf.square(diff) / (
            2 * tf.where(sigma_is_0, tf.ones_like(sigma_tens), sigma_tens) ** 2
        )
        norms = exp_term - tf.math.log(
            tf.where(
                sigma_is_0, tf.ones_like(sigma_tens), np.sqrt(2 * np.pi) * sigma_tens
            )
        )
        kernels = tf.reduce_sum(norms, -1)
        kernels = tf.exp(kernels)
        kernels /= tf.reduce_sum(kernels)
        kernels = tf.expand_dims(tf.expand_dims(kernels, -1), -1)

    return kernels


def sobel_kernels(n_dims):
    """Returns sobel kernels to compute spatial derivative on image of n dimensions."""

    in_dir = tf.convert_to_tensor([1, 0, -1], dtype="float32")
    orthogonal_dir = tf.convert_to_tensor([1, 2, 1], dtype="float32")
    comb = np.array(list(combinations(list(range(n_dims)), n_dims - 1))[::-1])

    list_kernels = list()
    for dim in range(n_dims):

        sublist_kernels = list()
        for axis in range(n_dims):

            kernel = in_dir if axis == dim else orthogonal_dir
            for i in comb[axis]:
                kernel = tf.expand_dims(kernel, axis=i)
            sublist_kernels.append(tf.expand_dims(tf.expand_dims(kernel, -1), -1))

        list_kernels.append(sublist_kernels)

    return list_kernels


def unit_kernel(dist_threshold, n_dims, max_dist_threshold=None):
    """Build kernel with values of 1 for voxel at a distance < dist_threshold from the center, and 0 otherwise.
    The outputs are given as tensorflow tensors.
    :param dist_threshold: maximum distance from the center until voxel will have a value of 1. Can be a tensor of size
    (batch_size, 1), or a float.
    :param n_dims: dimension of the kernel to return (excluding batch and channel dimensions).
    :param max_dist_threshold: if distance_threshold is a tensor, max_dist_threshold must be given. It represents the
    maximum value that will be passed to dist_threshold. Must be a float.
    """

    # convert dist_threshold into a tensor
    if not tf.is_tensor(dist_threshold):
        dist_threshold_tens = tf.convert_to_tensor(
            utils.reformat_to_list(dist_threshold), dtype="float32"
        )
    else:
        assert (
            max_dist_threshold is not None
        ), "max_sigma must be provided when dist_threshold is given as a tensor"
        dist_threshold_tens = tf.cast(dist_threshold, "float32")
    shape = dist_threshold_tens.get_shape().as_list()

    # get batchsize
    batchsize = (
        None
        if shape[0] is not None
        else tf.split(tf.shape(dist_threshold_tens), [1, -1])[0]
    )

    # set max_dist_threshold into an array
    if (
        max_dist_threshold is None
    ):  # dist_threshold is fixed (i.e. dist_threshold will not change at each mini-batch)
        max_dist_threshold = dist_threshold

    # get size of blurring kernels
    windowsize = np.array([max_dist_threshold * 2 + 1] * n_dims, dtype="int32")

    # build tensor representing the distance from the centre
    mesh = [
        tf.cast(f, "float32") for f in volshape_to_meshgrid(windowsize, indexing="ij")
    ]
    dist = tf.stack(
        [mesh[f] - (windowsize[f] - 1) / 2 for f in range(len(windowsize))], axis=-1
    )
    dist = tf.sqrt(tf.reduce_sum(tf.square(dist), axis=-1))

    # replicate distance to batch size and reshape sigma_tens
    if batchsize is not None:
        dist = tf.tile(
            tf.expand_dims(dist, axis=0),
            tf.concat(
                [batchsize, tf.ones(tf.shape(tf.shape(dist)), dtype="int32")], axis=0
            ),
        )
        for i in range(n_dims - 1):
            dist_threshold_tens = tf.expand_dims(dist_threshold_tens, axis=1)
    else:
        for i in range(n_dims - 1):
            dist_threshold_tens = tf.expand_dims(dist_threshold_tens, axis=0)

    # build final kernel by thresholding distance tensor
    kernel = tf.where(
        tf.less_equal(dist, dist_threshold_tens),
        tf.ones_like(dist),
        tf.zeros_like(dist),
    )
    kernel = tf.expand_dims(tf.expand_dims(kernel, -1), -1)

    return kernel


def resample_tensor(
    tensor,
    resample_shape,
    interp_method="linear",
    subsample_res=None,
    volume_res=None,
    build_reliability_map=False,
):
    """This function resamples a volume to resample_shape. It does not apply any pre-filtering.
    A prior downsampling step can be added if subsample_res is specified. In this case, volume_res should also be
    specified, in order to calculate the downsampling ratio. A reliability map can also be returned to indicate which
    slices were interpolated during resampling from the downsampled to final tensor.
    :param tensor: tensor
    :param resample_shape: list or numpy array of size (n_dims,)
    :param interp_method: (optional) interpolation method for resampling, 'linear' (default) or 'nearest'
    :param subsample_res: (optional) if not None, this triggers a downsampling of the volume, prior to the resampling
    step. List or numpy array of size (n_dims,). Default si None.
    :param volume_res: (optional) if subsample_res is not None, this should be provided to compute downsampling ratio.
    list or numpy array of size (n_dims,). Default is None.
    :param build_reliability_map: whether to return reliability map along with the resampled tensor. This map indicates
    which slices of the resampled tensor are interpolated (0=interpolated, 1=real slice, in between=degree of realness).
    :return: resampled volume, with reliability map if necessary.
    """

    # reformat resolutions to lists
    subsample_res = utils.reformat_to_list(subsample_res)
    volume_res = utils.reformat_to_list(volume_res)
    n_dims = len(resample_shape)

    # downsample image
    tensor_shape = tensor.get_shape().as_list()[1:-1]
    downsample_shape = tensor_shape  # will be modified if we actually downsample

    if subsample_res is not None:
        assert (
            volume_res is not None
        ), "volume_res must be given when providing a subsampling resolution."
        assert len(subsample_res) == len(volume_res), (
            "subsample_res and volume_res must have the same length, "
            "had {0}, and {1}".format(len(subsample_res), len(volume_res))
        )
        if subsample_res != volume_res:

            # get shape at which we downsample
            downsample_shape = [
                int(tensor_shape[i] * volume_res[i] / subsample_res[i])
                for i in range(n_dims)
            ]

            # downsample volume
            tensor._keras_shape = tuple(tensor.get_shape().as_list())
            tensor = nrn_layers.Resize(size=downsample_shape, interp_method="nearest")(
                tensor
            )

    # resample image at target resolution
    if (
        resample_shape != downsample_shape
    ):  # if we didn't downsample downsample_shape = tensor_shape
        tensor._keras_shape = tuple(tensor.get_shape().as_list())
        tensor = nrn_layers.Resize(size=resample_shape, interp_method=interp_method)(
            tensor
        )

    # compute reliability maps if necessary and return results
    if build_reliability_map:

        # compute maps only if we downsampled
        if downsample_shape != tensor_shape:

            # compute upsampling factors
            upsampling_factors = np.array(resample_shape) / np.array(downsample_shape)

            # build reliability map
            reliability_map = 1
            for i in range(n_dims):
                loc_float = np.arange(0, resample_shape[i], upsampling_factors[i])
                loc_floor = np.int32(np.floor(loc_float))
                loc_ceil = np.int32(np.clip(loc_floor + 1, 0, resample_shape[i] - 1))
                tmp_reliability_map = np.zeros(resample_shape[i])
                tmp_reliability_map[loc_floor] = 1 - (loc_float - loc_floor)
                tmp_reliability_map[loc_ceil] = tmp_reliability_map[loc_ceil] + (
                    loc_float - loc_floor
                )
                shape = [1, 1, 1]
                shape[i] = resample_shape[i]
                reliability_map = reliability_map * np.reshape(
                    tmp_reliability_map, shape
                )
            shape = KL.Lambda(lambda x: tf.shape(x))(tensor)
            mask = KL.Lambda(
                lambda x: tf.reshape(
                    tf.convert_to_tensor(reliability_map, dtype="float32"), shape=x
                )
            )(shape)

        # otherwise just return an all-one tensor
        else:
            mask = KL.Lambda(lambda x: tf.ones_like(x))(tensor)

        return tensor, mask

    else:
        return tensor


def expand_dims(tensor, axis=0):
    """Expand the dimensions of the input tensor along the provided axes (given as an integer or a list)."""
    axis = utils.reformat_to_list(axis)
    for ax in axis:
        tensor = tf.expand_dims(tensor, axis=ax)
    return tensor
